﻿using System;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Collections.Generic;

using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;

using Platform.Extensions;
using Platform.Configuration.DatabaseConfiguration;
using Platform.Modules.Abstract;
using Platform.Logging;
using Platform.Configuration;

namespace Platform.Modules.Implementation
{
    public class ModuleHost : IModuleHost
    {
        #region Construction and Fields

        // the list of registered modules
        protected readonly List<IModule> _Modules = new List<IModule>();

        // the list of command line arguments
        protected string[] _Args;

        /// <summary>
        /// Construct a new module host
        /// </summary>
        /// <param name="name"></param>
        public ModuleHost( string name, params string[] args )
        {
            // initialize properties
            Id = Guid.NewGuid().ToString().Replace("-", "").ToUpper();
            Folder = Assembly.GetExecutingAssembly().GetFolder();
            Name = name;
            _Args = args;

            // dynamic updates of default settings
            DefaultSettings["Logging:File:Path"] = Path.Combine("c:\\logging\\", name.MakeValidFilePath());
            DefaultSettings["Logging:File:Filename"] = name.MakeValidFileName();
        }

        #endregion

        #region Properties

        /// <summary>
        /// Gets the id
        /// </summary>
        public string Id { get; }

        /// <summary>
        /// Gets the name
        /// </summary>
        public string Name { get; private set; }

        /// <summary>
        /// Gets the application folder
        /// </summary>
        public string Folder { get; }

        /// <summary>
        /// Gets the host instance name (ex. physical location or hostname)
        /// </summary>
        public string Instance { get; protected set; }

        /// <summary>
        /// Gets the configuration
        /// </summary>
        public IConfigurationRoot Configuration { get; protected set;  }

        /// <summary>
        /// Gets the services
        /// </summary>
        public IServiceCollection Services { get; protected set; }

        /// <summary>
        /// Gets the provider
        /// </summary>
        public IServiceProvider Provider { get; protected set; }

        /// <summary>
        /// Default configuraion settings
        /// </summary>
        public Dictionary<string,string> DefaultSettings = new Dictionary<string, string>
            {
                {"Logging:IncludeScopes", "false" },
                {"Logging:LogLevel:Default", "Information"},
                {"Logging:Path", "c:\\logging" },
                {"Logging:Http:IsEnabled", "false" },
                {"Logging:Database:IsEnabled", "false" },
                {"Logging:Debug:IsEnabled", "true" },
                {"Logging:Console:IsEnabled", "true" },
                {"Logging:Interactive:IsEnabled", "false" },
                {"Logging:File:IsEnabled", "true" }
            };

        #endregion

        #region Events

        /// <summary>
        /// Signals host initialization complete
        /// </summary>
        // public event EventHandler Initialized;

        #endregion

        #region Initialize Host and Modules

        public Action<IConfigurationBuilder> BuildConfigurationCallback { get; set; }

        protected virtual void BuildConfiguration( IConfigurationBuilder builder )
        {
            builder.SetBasePath(Folder)
                .AddInMemoryCollection(DefaultSettings)
                .AddJsonFile((Name + ".json").MakeValidFilePath(), optional: true, reloadOnChange: true)
                .AddDatabaseConfiguration()
                .AddEnvironmentVariables()
                .AddCommandLine(_Args);

            BuildConfigurationCallback?.Invoke(builder);
        }

        public Action<IServiceCollection> BuildServicesCallback { get; set; }

        protected virtual void BuildServices(IServiceCollection builder)
        {
            builder.AddOptions();
            builder.AddLogging(BuildLogging);
            builder.AddSingleton<IModuleHost>(this);

            // initialize all modules
            foreach (var module in _Modules)
                module.Initialize(this, builder);

            BuildServicesCallback?.Invoke(builder);
        }

        public Action<ILoggingBuilder> BuildLoggingCallback { get; set; }

        protected virtual void BuildLogging( ILoggingBuilder builder )
        {
            BuildLoggingCallback?.Invoke(builder);

            // default logging
            var section = Configuration.GetSection("Logging");
            builder.AddConfiguration(section);
            builder.AddPlatformConsoleLogger(section);
            builder.AddPlatformFileLogger(section);
            builder.AddPlatformDatabaseLogger(section);
            builder.AddPlatformHttpLogger(section);
            builder.AddPlatformInteractiveLogger(section);
            builder.AddDebug();

            // enable custom logging
            BuildLoggingCallback?.Invoke(builder);
        }


        /// <summary>
        /// Initialize all modules
        /// </summary>
        public virtual void Initialize()
        {
            // build configuration
            var configuration = new ConfigurationBuilder();
            BuildConfiguration(configuration);
            Configuration = configuration.Build();

            // get the instance name
            Instance = Configuration["InstanceName"] ?? Environment.MachineName;

            // Create services collection
            var services = new ServiceCollection();
            BuildServices(services);
            Provider = services.BuildServiceProvider();

            // initialize non DI logger provider
            LoggingProvider.Initialize(Provider);
            // initialize non DI service provider
            ServiceProvider.Initialize(Provider);
        }

        #endregion

        #region Modules

        /// <summary>
        /// Add module
        /// </summary>
        /// <param name="module"></param>
        public void AddModule( IModule module )
        {
            _Modules.Add(module);
        }

        /// <summary>
        /// Gets a module by name
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        public IModule GetModule(string name)
        {
            return _Modules.FirstOrDefault(m => m.Name == name);
        }

        /// <summary>
        /// Restart Services
        /// </summary>
        public void Restart()
        {
            Stop();
            Start();
        }

        /// <summary>
        /// Start all modules
        /// </summary>
        public void Start()
        {
            // start all modules
            foreach (var module in _Modules)
                module.Start();

            // start all tasks
            foreach (var task in Provider.GetServices<IModuleTask>())
                task.Start();
        }

        /// <summary>
        /// Stop all modules
        /// </summary>
        public void Stop()
        {
            // stop all tasks
            foreach (var task in Provider.GetServices<IModuleTask>())
                task.Stop();

            // stop all modules
            foreach (var module in _Modules)
                module.Stop();
        }

        #endregion
    }
}
